from types import Currency

class Split(object):
    def __init__(self, expenditure=None, user=None, share=Currency(0)):
        self.expenditure = expenditure
        self.user = user
        self.share = share
    
    def __repr__(self):
        return '<Split: expense: %s user: %s share: %s>' % (self.expenditure,
                                                            self.user,
                                                            self.share)

__all__ = ['Split']



from types import Currency

class Transfer(object):
    def __init__(self, debtor=None, creditor=None, amount=Currency(0)):
        self.debtor = debtor
        self.creditor = creditor
        self.amount = amount
    
    def __repr__(self):
        return '<Transfer: from %s to %s for %s>' % (self.debtor,
                                                     self.creditor,
                                                     self.amount)

    def involves(self, user):
        return (self.debtor == user) or (self.creditor == user)

__all__ = ['Transfer']



"""
Calculate the total state of the books
"""

from bluechips import model
from bluechips.model import meta

from bluechips.model.types import Currency

import sqlalchemy

class DirtyBooks(Exception):
    """
    If the books don't work out, raise this
    """
    pass

def debts():
    # In this scheme, negative numbers represent money the house owes
    # the user, and positive numbers represent money the user owes the
    # house
    users = meta.Session.query(model.User)
    
    debts_dict = dict((u, Currency(0)) for u in users)
    
    # First, credit everyone for expenditures they've made
    total_expenditures = meta.Session.query(model.Expenditure).\
        add_column(sqlalchemy.func.sum(model.Expenditure.amount).label('total_spend')).\
        group_by(model.Expenditure.spender_id)
    for expenditure, total_spend in total_expenditures:
        debts_dict[expenditure.spender] -= total_spend
    
    # Next, debit everyone for expenditures that they have an
    # investment in (i.e. splits)
    
    total_splits = meta.Session.query(model.Split).\
        add_column(sqlalchemy.func.sum(model.Split.share).label('total_split')).\
        group_by(model.Split.user_id)
    
    for split, total_cents in total_splits:
        debts_dict[split.user] += total_cents
    
    # Finally, move transfers around appropriately
    #
    # To keep this from getting to be expensive, have SQL sum up
    # transfers for us
    
    transfer_q = meta.Session.query(model.Transfer).\
        add_column(sqlalchemy.func.sum(model.Transfer.amount).label('total_amount'))
    total_debits = transfer_q.group_by(model.Transfer.debtor_id)
    total_credits = transfer_q.group_by(model.Transfer.creditor_id)
    
    for transfer, total_amount in total_debits:
        debts_dict[transfer.debtor] -= total_amount
    for transfer, total_amount in total_credits:
        debts_dict[transfer.creditor] += total_amount
    
    return debts_dict

def settle(debts_dict):
    # This algorithm has been shamelessly stolen from Nelson Elhage's
    # <nelhage@mit.edu> implementation for our 2008 summer apartment.
    
    debts_list = [dict(who=user, amount=amount) for user, amount in \
                      debts_dict.iteritems()]
    #debts_list.sort(reverse=True, key=(lambda x: abs(x['amount'])))
    
    owes_list = [debt for debt in debts_list if debt['amount'] > 0]
    owed_list = [debt for debt in debts_list if debt['amount'] < 0]
    
    settle_list = []
    
    while len(owes_list) > 0 and len(owed_list) > 0:
        owes_list.sort(reverse=True, key=(lambda x: abs(x['amount'])))
        owed_list.sort(reverse=True, key=(lambda x: abs(x['amount'])))

        owes = owes_list[0]
        owed = owed_list[0]
        
        sum = owes['amount'] + owed['amount']
        if sum == 0:
            # Perfect balance!
            owes_list.pop(0)
            owed_list.pop(0)
            val = owes['amount']
        elif sum > 0:
            # person in owes still owes money
            owes['amount'] += owed['amount']
            owed_list.pop(0)
            val = -owed['amount']
        else:
            # person in owed is owed more than owes has to give
            owed['amount'] += owes['amount']
            owes_list.pop(0)
            val = owes['amount']
        
        settle_list.append((owes['who'], owed['who'], val))
    
    if len(owes_list) > 0:
        raise DirtyBooks, ("People still owe money", owes_list)
    if len(owed_list) > 0:
        raise DirtyBooks, ("People are still owed money", owed_list)
    
    return settle_list

__all__ = ['debts', 'settle']



"""The base Controller API

Provides the BaseController class for subclassing.
"""

from decorator import decorator

from pylons import request, session, tmpl_context as c
from pylons.controllers import WSGIController
from pylons.i18n import _, ungettext, N_
from pylons.templating import render_mako

from mako.exceptions import TopLevelLookupException

import bluechips.lib.helpers as h
from bluechips import model
from bluechips.model import meta


class BaseController(WSGIController):

    def __call__(self, environ, start_response):
        """Invoke the Controller"""
        # WSGIController.__call__ dispatches to the Controller method
        # the request is routed to. This routing information is
        # available in environ['pylons.routes_dict']
        try:
            return WSGIController.__call__(self, environ, start_response)
        finally:
            meta.Session.remove()

def update_sar(record, form_result):
    """
    Update a SQLAlchemy record with the results of a validated form submission
    """
    for key, value in form_result.items():
        setattr(record, key, value)

def redirect_on_get(action):
    """
    Decorator for a controller action. If the action is called with a GET
    method, 302 redirect to the action specified.
    """

    @decorator
    def redirect_on_get_wrap(func, *args, **kwargs):
        if request.method == 'GET':
            controller = request.environ['pylons.routes_dict']['controller']
            return h.redirect_to(controller=controller, action=action)
        else:
            return func(*args, **kwargs)
    return redirect_on_get_wrap

def render(name, *args, **kwargs):
    if any([x in request.user_agent for x in ('iPhone','webOS')]):
        if 'use_non_mobile' in request.params:
            session['use_non_mobile'] = (request.params['use_non_mobile'] ==
                                         'yes')
        if session.get('use_non_mobile'):
            c.mobile_client = True
        else:
            try:
                return render_mako('/mobile' + name, *args, **kwargs)
            except TopLevelLookupException:
                # If a mobile template doesn't exist for this page, don't show
                # the 'use mobile interface' link.
                c.mobile_client = False
    return render_mako(name, *args, **kwargs)

def get_users():
    return meta.Session.query(model.User.id, model.User).\
        order_by(model.User.resident.desc(), model.User.username)

__all__ = ['c', 'h', 'render', 'model', 'meta', '_', 'ungettext', 'N_',
           'BaseController', 'update_sar', 'redirect_on_get', 'get_users']



from bluechips.tests import *

class TestMobileController(TestController):
    def setUp(self):
        self.ua = ('Mozilla/5.0 (iPhone; U; CPU like Mac OS X; en) '
                   'AppleWebKit/420+ (KHTML, like Gecko) Version/3.0 '
                   'Mobile/1A543a Safari/419.3')
        self.app.extra_environ['HTTP_USER_AGENT'] = self.ua

    def test_switch_interfaces(self):
        response = self.app.get('/')
        response.mustcontain('apple-touch-icon')
        response.mustcontain('Use non mobile interface')
        response = response.click('Use non mobile interface')
        response.mustcontain('Use mobile interface')
        response = response.click('Use mobile interface')

    def test_view_nonmobile(self):
        self.app.get(url_for(controller='history'))



"""Helper functions

Consists of functions to typically be used within templates, but also
available to Controllers. This module is available to both as 'h'.
"""
from datetime import date
from decimal import Decimal

from pylons import request
from routes import url_for, redirect_to
from webhelpers.html import escape, literal, url_escape
from webhelpers.html.tags import *
from webhelpers.pylonslib.secure_form import *

from webhelpers.pylonslib import Flash as _Flash


def currency(name, value, *args, **kwargs):
    if 'class_' not in kwargs:
        kwargs['class_'] = ''
    kwargs['class_'] += 'currency'
    value = "%0.2f" % (int(value) / 100.)
    return text(name, value, *args, **kwargs)


def grab(obj, attr):
    if obj:
        return getattr(obj, attr)
    else:
        if attr == 'date':
            return date.today()
        elif attr in ('spender_id', 'creditor_id', 'debtor_id'):
            return request.environ['user'].id
        elif attr == 'amount':
            return 0
        else:
            return ''

flash = _Flash()



"""
Handle transfers
"""

import logging

from datetime import date

from bluechips.lib.base import *

from pylons import request, app_globals as g
from pylons.decorators import validate
from pylons.decorators.secure import authenticate_form
from pylons.controllers.util import abort

from formencode import Schema, validators

from mailer import Message

log = logging.getLogger(__name__)


class TransferSchema(Schema):
    "Validate a transfer."
    allow_extra_fields = False
    debtor_id = validators.Int(not_empty=True)
    creditor_id = validators.Int(not_empty=True)
    amount = model.types.CurrencyValidator(not_empty=True)
    description = validators.UnicodeString()
    date = validators.DateConverter()
 

class TransferController(BaseController):
    def index(self):
       return self.edit()
    
    def edit(self, id=None):
        c.users = get_users()
        if id is None:
            c.title = 'Add a New Transfer'
            c.transfer = model.Transfer()
            c.transfer.debtor_id = request.environ['user'].id
            c.transfer.date = date.today()
        else:
            c.title = 'Edit a Transfer'
            c.transfer = meta.Session.query(model.Transfer).get(id)
            if c.transfer is None:
                abort(404)
        return render('/transfer/index.mako')
    
    @redirect_on_get('edit')
    @authenticate_form
    @validate(schema=TransferSchema(), form='edit')
    def update(self, id=None):
        if id is None:
            t = model.Transfer()
            meta.Session.add(t)
            op = 'created'
        else:
            t = meta.Session.query(model.Transfer).get(id)
            if t is None:
                abort(404)
            op = 'updated'
        
        update_sar(t, self.form_result)
        meta.Session.commit()
       
        show = ('Transfer of %s from %s to %s %s.' %
                (t.amount, t.debtor, t.creditor, op))
        h.flash(show)

        # Send email notification to involved users if they have an email set.
        body = render('/emails/transfer.txt', extra_vars={'transfer': t,
                                                          'op': op})
        g.handle_notification((t.debtor, t.creditor), show, body)

        return h.redirect_to('/')

    def delete(self, id):
        c.title = 'Delete a Transfer'
        c.transfer = meta.Session.query(model.Transfer).get(id)
        if c.transfer is None:
            abort(404)

        return render('/transfer/delete.mako')

    @redirect_on_get('delete')
    @authenticate_form
    def destroy(self, id):
        t = meta.Session.query(model.Transfer).get(id)
        if t is None:
            abort(404)

        if 'delete' in request.params:
            meta.Session.delete(t)

            meta.Session.commit()
            show = ("Transfer of %s from %s to %s deleted." %
                    (t.amount, t.debtor, t.creditor))
            h.flash(show)

            body = render('/emails/transfer.txt',
                          extra_vars={'transfer': t,
                                      'op': 'deleted'})
            g.handle_notification((t.debtor, t.creditor), show, body)

        return h.redirect_to('/')



from unittest import TestCase
from bluechips.model import types

class TestCurrency(TestCase):
    def setUp(self):
        self.c = types.Currency('12.34')

    def test_currency_float(self):
        assert float(self.c) == 1234.

    def test_currency_int(self):
        val = int(self.c)
        assert val == 1234
        assert type(val) == int

    def test_currency_long(self):
        val = long(self.c)
        assert val == 1234
        assert type(val) == long



from unittest import TestCase
from bluechips.tests import *
from bluechips import model
from bluechips.model import meta
from bluechips.model.types import Currency
from decimal import Decimal

class TestSplitFixed(TestCase):
    def test_simpleSplit(self):
        """
        Test simply splitting a $100 expenditure amongst 4 people
        """
        createUsers(4)
        
        e = model.Expenditure(meta.Session.query(model.User).first(),
                              Currency("100"))
        meta.Session.add(e)
        e.even_split()
        meta.Session.commit()
        
        for s in meta.Session.query(model.Split).\
                filter(model.Split.expenditure==e):
            self.assertEqual(s.share, Currency("25.00"))
        
        deleteExpenditures()
        deleteUsers()
    
    def test_uneven(self):
        """
        Test that expenditures can be split non-evenly
        """
        createUsers(2)
        
        users = meta.Session.query(model.User).all()
        
        e = model.Expenditure(users[0], Currency("100"))
        meta.Session.add(e)
        
        split_dict = {users[0]: Decimal("20"),
                      users[1]: Decimal("80")}
        
        amount_dict = {users[0]: Currency("20"),
                       users[1]: Currency("80")}
        
        e.split(split_dict)
        meta.Session.commit()
        
        for s in meta.Session.query(model.Split):
            self.assertEqual(s.share, amount_dict[s.user])
        
        deleteExpenditures()
        deleteUsers()
    
    def test_unevenBadTotal(self):
        """
        Test that transactions get split up properly when the uneven
        split shares don't add to 100%
        """
        createUsers(2)
        
        users = meta.Session.query(model.User).all()
        
        e = model.Expenditure(users[0], Currency("100.00"))
        meta.Session.add(e)
        
        split_dict = {users[0]: Decimal(10),
                      users[1]: Decimal(15)}
        
        amount_dict = {users[0]: Currency("40"),
                       users[1]: Currency("60")}
        
        e.split(split_dict)
        meta.Session.commit()
        
        for s in meta.Session.query(model.Split):
            self.assertEqual(s.share, amount_dict[s.user])
        
        deleteExpenditures()
        deleteUsers()

    def test_negativeExpenditure(self):
        """
        Test that negative expenditures get split correctly
        """
        createUsers(2)

        users = meta.Session.query(model.User).all()

        e = model.Expenditure(users[0], Currency("100.00"))
        meta.Session.add(e)

        # Force a split that will result in needing to distribute
        # pennies
        split_dict = {users[0]: Decimal(1),
                      users[1]: Decimal(2)}
        e.split(split_dict)
        meta.Session.commit()

        self.assertEqual(e.amount, sum(s.share for s in meta.Session.query(model.Split)))

        deleteExpenditures()
        deleteUsers()



"""Pylons application test package

This package assumes the Pylons environment is already loaded, such as
when this script is imported from the `nosetests --with-pylons=test.ini`
command.

This module initializes the application via ``websetup`` (`paster
setup-app`) and provides the base testing objects.
"""
from unittest import TestCase

from paste.deploy import loadapp
from paste.fixture import TestApp
from paste.script.appinstall import SetupCommand
from pylons import config
from routes import url_for

import bluechips.model
from bluechips.model import meta
from bluechips.model.types import Currency

import random

__all__ = ['url_for', 'TestController',
           'createUsers', 'createExpenditures',
           'deleteUsers', 'deleteExpenditures']

sample_users = [u'Alice', u'Bob', u'Charlie', u'Dave', u'Eve']

def setUpPackage():
    # Invoke websetup with the current config file
    SetupCommand('setup-app').run([config['__file__']])
    
    u1 = bluechips.model.User(u'root', u'Charlie Root', True)
    u1.email = u'charlie@example.com'
    u1.password = u'charliepass'
    u2 = bluechips.model.User(u'ben', u'Ben Bitdiddle', True)
    u3 = bluechips.model.User(u'gotta', u'Gotta Lisp', True)
    u4 = bluechips.model.User(u'rich', u'Rich Scheme', True)

    for u in (u1, u2, u3, u4):
        meta.Session.add(u)
    meta.Session.commit()

def tearDownPackage():
    meta.metadata.drop_all()

class TestController(TestCase):

    def __init__(self, *args, **kwargs):
        wsgiapp = loadapp('config:%s' % config['__file__'])
        self.app = TestApp(wsgiapp)
        TestCase.__init__(self, *args, **kwargs)

def createUsers(n=None):
    if n is None:
        n = random.randint(2, 5)
    for i in xrange(n):
        u = bluechips.model.User(sample_users[i].lower(), resident=True)
        meta.Session.add(u)
    meta.Session.commit()

def createExpenditures(n=None):
    if n is None:
        n = random.randint(5, 20)
    users = meta.Session.query(bluechips.model.User).all()
    for i in xrange(n):
        e = bluechips.model.Expenditure(random.choice(users),
                                        Currency(random.randint(1000, 100000)))
        meta.Session.add(e)
        e.even_split()
    meta.Session.commit()

def deleteUsers():
    map(meta.Session.delete, meta.Session.query(bluechips.model.User))

def deleteExpenditures():
    map(meta.Session.delete, meta.Session.query(bluechips.model.Expenditure))



class User(object):
    def __init__(self, username, name=u"", resident=True):
        self.username = username
        self.name = name
        self.resident = resident
    
    def __repr__(self):
        return '<User: %s>' % (self.username)

    def __str__(self):
        return self.name

__all__ = ['User']



"""
Display old transactions
"""

import logging

from bluechips.lib.base import *
from bluechips.lib.totals import *

import sqlalchemy
from sqlalchemy import orm

log = logging.getLogger(__name__)

class HistoryController(BaseController):
    def index(self):
        c.title = 'History'
        
        c.expenditures = meta.Session.query(model.Expenditure).\
                options(orm.eagerload('splits')).all()
        c.transfers = meta.Session.query(model.Transfer).all()

        return render('/history/index.mako')



from unittest import TestCase
from bluechips import model

class TestTransfer(TestCase):
    def setUp(self):
        self.u1 = model.User('chaz', u'Charles Root', False)
        self.u2 = model.User('boo', u'Boo Ghost', True)
        self.t = model.Transfer(self.u1, self.u2, 1234)

    def test_constructor(self):
        assert self.t.debtor == self.u1
        assert self.t.creditor == self.u2
        assert self.t.amount == 1234

    def test_repr(self):
        assert (repr(self.t) == 
                '<Transfer: from Charles Root to Boo Ghost for 1234>')

    def test_involves(self):
        other_u = model.User('jim', u'Jimbo James', True)
        assert self.t.involves(other_u) == False
        assert self.t.involves(self.u1)
        assert self.t.involves(self.u2)



from unittest import TestCase
from bluechips.lib import permissions

class TestReorderingSettle(TestCase):
    def test_authenticate(self):
        assert permissions.authenticate({}, u'root', u'charliepass')
        assert not permissions.authenticate({}, u'root', u'blah')
        assert not permissions.authenticate({}, u'blah', u'charliepass')
        assert not permissions.authenticate({}, u'blah', u'blah')



"""
Handle expenditures
"""

import logging

from decimal import Decimal, InvalidOperation

from bluechips.lib.base import *

from pylons import request, app_globals as g
from pylons.decorators import validate
from pylons.decorators.secure import authenticate_form
from pylons.controllers.util import abort

from formencode import validators, Schema
from formencode.foreach import ForEach
from formencode.variabledecode import NestedVariables
from formencode.schema import SimpleFormValidator

from mailer import Message

log = logging.getLogger(__name__)


class ShareSchema(Schema):
    "Validate individual user shares."
    allow_extra_fields = False
    user_id = validators.Int(not_empty=True)
    amount = validators.Number(not_empty=True)


def validate_state(value_dict, state, validator):
    if all(s['amount'] == 0 for s in value_dict['shares']):
        return {'shares-0.amount': 'Need at least one non-zero share'}
ValidateNotAllZero = SimpleFormValidator(validate_state)


class ExpenditureSchema(Schema):
    "Validate an expenditure."
    allow_extra_fields = False
    pre_validators = [NestedVariables()]
    spender_id = validators.Int(not_empty=True)
    amount = model.types.CurrencyValidator(not_empty=True)
    description = validators.UnicodeString(not_empty=True)
    date = validators.DateConverter()
    shares = ForEach(ShareSchema)
    chained_validators = [ValidateNotAllZero]
    

class SpendController(BaseController):
    def index(self):
        return self.edit()
    
    def edit(self, id=None):
        c.users = get_users()
        if id is None:
            c.title = 'Add a New Expenditure'
            c.expenditure = model.Expenditure()
            c.expenditure.spender_id = request.environ['user'].id

            num_residents = meta.Session.query(model.User).\
                    filter_by(resident=True).count()
            # Pre-populate split percentages for an even split.
            c.values = {}
            for ii, user_row in enumerate(c.users):
                user_id, user = user_row
                val = 0
                if user.resident:
                    val = Decimal(100) / Decimal(num_residents)
                c.values['shares-%d.amount' % ii] = val
        else:
            c.title = 'Edit an Expenditure'
            c.expenditure = meta.Session.query(model.Expenditure).get(id)
            if c.expenditure is None:
                abort(404)
            c.values = {}
            for ii, user_row in enumerate(c.users):
                user_id, user = user_row
                shares_by_user = dict(((sp.user, sp.share) for sp
                                       in c.expenditure.splits))
                share = shares_by_user.get(user, 0)
                if c.expenditure.amount == 0:
                    percent = 0
                else:
                    percent = (Decimal(100) * Decimal(int(share)) /
                               Decimal(int(c.expenditure.amount))).\
                            quantize(Decimal("0.001"))
                c.values['shares-%d.amount' % ii] = percent

        return render('/spend/index.mako')

    @redirect_on_get('edit')
    @authenticate_form
    @validate(schema=ExpenditureSchema(), form='edit', variable_decode=True)
    def update(self, id=None):
        # Either create a new object, or, if we're editing, get the
        # old one
        if id is None:
            e = model.Expenditure()
            meta.Session.add(e)
            op = 'created'
        else:
            e = meta.Session.query(model.Expenditure).get(id)
            if e is None:
                abort(404)
            op = 'updated'
        
        # Set the fields that were submitted
        shares = self.form_result.pop('shares')
        update_sar(e, self.form_result)

        users = dict(meta.Session.query(model.User.id, model.User).all())
        split_dict = {}
        for share_params in shares:
            user = users[share_params['user_id']]
            split_dict[user] = Decimal(str(share_params['amount']))
        e.split(split_dict)
        
        meta.Session.commit()
       
        show = ("Expenditure of %s paid for by %s %s." %
                (e.amount, e.spender, op))
        h.flash(show)

        # Send email notification to involved users if they have an email set.
        involved_users = set(sp.user for sp in e.splits if sp.share != 0)
        involved_users.add(e.spender)
        body = render('/emails/expenditure.txt',
                      extra_vars={'expenditure': e,
                                  'op': op})
        g.handle_notification(involved_users, show, body)

        return h.redirect_to('/')

    def delete(self, id):
        c.title = 'Delete an Expenditure'
        c.expenditure = meta.Session.query(model.Expenditure).get(id)
        if c.expenditure is None:
            abort(404)

        return render('/spend/delete.mako')

    @redirect_on_get('delete')
    @authenticate_form
    def destroy(self, id):
        e = meta.Session.query(model.Expenditure).get(id)
        if e is None:
            abort(404)

        if 'delete' in request.params:
            meta.Session.delete(e)

            meta.Session.commit()
            show = ("Expenditure of %s paid for by %s deleted." %
                    (e.amount, e.spender))
            h.flash(show)

            involved_users = set(sp.user for sp in e.splits if sp.share != 0)
            involved_users.add(e.spender)
            body = render('/emails/expenditure.txt',
                          extra_vars={'expenditure': e,
                                      'op': 'deleted'})
            g.handle_notification(involved_users, show, body)

        return h.redirect_to('/')



import cgi

from pylons import request, tmpl_context as c

from bluechips.lib.base import BaseController, render

class ErrorController(BaseController):
    """Generates error documents as and when they are required.

    The ErrorDocuments middleware forwards to ErrorController when error
    related status codes are returned from the application.

    This behaviour can be altered by changing the parameters to the
    ErrorDocuments middleware in your config/middleware.py file.
    
    """
    def document(self):
        """Render the error document"""
        resp = request.environ.get('pylons.original_response')
        c.code = cgi.escape(request.GET.get('code', resp.status))
        return render('/error.mako')



from decimal import Decimal

from unittest import TestCase
from bluechips import model
from bluechips.model import meta
from bluechips.model.types import Currency

class TestExpenditure(TestCase):
    def setUp(self):
        self.u = model.User(u'chaz', u'Charles Root', False)
        self.e = model.Expenditure(self.u, Currency('444.88'),
                                   u'chaz buys lunch')
        meta.Session.add(self.u)
        meta.Session.add(self.e)
        meta.Session.commit()

    def test_constructor(self):
        assert self.e.spender == self.u
        assert self.e.amount == Currency('444.88')
        assert self.e.description == u'chaz buys lunch'

    def test_repr(self):
        assert (repr(self.e) == 
                '<Expenditure: spender: Charles Root spent: $444.88>')

    def test_even_split(self):
        self.e.even_split()
        meta.Session.commit()
        for sp in self.e.splits:
            assert sp.share == Currency('111.22')

    def test_split_change_to_zero(self):
        self.e.even_split()
        meta.Session.commit()
        users = meta.Session.query(model.User).all()
        split_dict = dict((user, Decimal('0')) for user in users)
        split_dict[self.u] = Decimal(1)
        self.e.split(split_dict)

    def _two_way_split_test(self, amount, min, max):
        e2 = model.Expenditure(self.u, amount,
                              u'testing splits')
        u2 = model.User(u'bo', u'Bo Jangles', False)
        meta.Session.add(u2)
        meta.Session.add(e2)
        meta.Session.commit()
        split_dict = {}
        split_dict[self.u] = Decimal(1)
        split_dict[u2] = Decimal(1)
        e2.split(split_dict)
        assert min <= e2.share(u2) <= max
        meta.Session.delete(e2)
        meta.Session.delete(u2)
        meta.Session.commit()

    def test_split_rounds_down(self):
        self._two_way_split_test(Currency('40.01'),
                                 Currency('20.00'),
                                 Currency('20.01'))

    def test_split_rounds_up(self):
        self._two_way_split_test(Currency('39.99'),
                                 Currency('19.99'),
                                 Currency('20.00'))

    def test_split_small(self):
        self._two_way_split_test(Currency('0.01'),
                                 Currency('0.00'),
                                 Currency('0.01'))

    def test_split_small_negative(self):
        self._two_way_split_test(Currency('-0.01'),
                                 Currency('-0.01'),
                                 Currency('-0.00'))

    def test_split_irrational_rounding(self):
        e2 = model.Expenditure(self.u, Decimal('2375.00'),
                               u'rounding test')
        u2 = model.User(u'rat', u'Irrational Rat', False)
        meta.Session.add(u2)
        meta.Session.add(e2)
        meta.Session.commit()
        split_dict = {}
        split_dict[u2] = Decimal('750.00')
        split_dict[self.u] = Decimal('4000.00')
        e2.split(split_dict)
        assert e2.share(u2) == Decimal('375.00')
        meta.Session.delete(e2)
        meta.Session.delete(u2)
        meta.Session.commit()

    def tearDown(self):
        meta.Session.delete(self.e)
        meta.Session.delete(self.u)
        meta.Session.commit()



"""
authkit authorization permission objects for BlueChips
"""

from authkit.authenticate import AddDictToEnviron
from authkit.authorize import NotAuthenticatedError, NotAuthorizedError
from authkit.permissions import RequestPermission

from bluechips import model
from bluechips.model import meta

class BlueChipUser(RequestPermission):
    def check(self, app, environ, start_response):
        if 'REMOTE_USER' not in environ:
            raise NotAuthenticatedError('Not Authenticated') # pragma: nocover
        environ['user'] = meta.Session.query(model.User).\
            filter_by(username=unicode(environ['REMOTE_USER'])).\
            first()
        if environ['user'] == None:
            raise NotAuthorizedError('You are not allowed access.') # pragma: nocover
        return app(environ, start_response)

class BlueChipResident(RequestPermission):
    def check(self, app, environ, start_response):
        if 'user' not in environ:
            raise NotAuthenticatedError('Not Authenticated')

        if not getattr(environ['user'], 'resident', False):
            raise NotAuthorizedError('You are not allowed access.')

        return app(environ, start_response)

class DummyAuthenticate(AddDictToEnviron):
    """
    Set the authkit.authenticate environment variable so
    authkit.authorize shuts up
    """
    def __init__(self, app, app_conf):
        newenv = {}
        newenv['authkit.authenticate'] = True
        newenv['authkit.config'] = {'setup.enable': True}
        if 'fake_username' in app_conf:
            newenv['REMOTE_USER'] = app_conf['fake_username']
        super(DummyAuthenticate, self).__init__(app, newenv)


def authenticate(environ, username, password):
    user = meta.Session.query(model.User).\
            filter_by(username=unicode(username),
                      password=unicode(password)).first()
    return (user is not None)

__all__ = ['BlueChipUser', 'DummyAuthenticate', 'authenticate']



from unittest import TestCase
from bluechips import model

class TestUser(TestCase):
    def setUp(self):
        self.u = model.User('chaz', u'Charles Root', False)

    def test_constructor(self):
        assert self.u.username == 'chaz'
        assert self.u.name == u'Charles Root'
        assert self.u.resident == False

    def test_repr(self):
        assert repr(self.u) == '<User: chaz>'

    def test_str(self):
        assert str(self.u) == 'Charles Root'



from unittest import TestCase
from formencode import Invalid

from bluechips.model import types

class TestCurrencyValidator(TestCase):
    def setUp(self):
        self.v = types.CurrencyValidator()

    def test_currency_validator_good(self):
        assert (self.v.to_python('12.34') ==
                types.Currency('12.34'))

    def test_currency_validator_nonzero(self):
        try:
            self.v.to_python('0')
        except Invalid:
            pass

    def test_currency_validator_precision(self):
        try:
            self.v.to_python('12.345')
        except Invalid:
            pass

    def test_currency_validator_amount(self):
        try:
            self.v.to_python('foo')
        except Invalid:
            pass



from unittest import TestCase
from bluechips import model
from bluechips.model.types import Currency

class TestSplit(TestCase):
    def setUp(self):
        self.u = model.User('chaz', u'Charles Root', False)
        self.e = model.Expenditure(self.u, Currency('12.34'),
                                   u'A test expenditure')
        self.sp = model.Split(self.e, self.u, Currency('5.55'))

    def test_constructor(self):
        assert self.sp.expenditure == self.e
        assert self.sp.user == self.u
        assert self.sp.share == Currency('5.55')

    def test_repr(self):
        assert (repr(self.sp) == '<Split: expense: %s user: %s share: %s>' %
                (self.sp.expenditure, self.sp.user, self.sp.share))



from unittest import TestCase
from pylons import request
from datetime import date
from bluechips.lib import helpers as h

class TestHelpers(TestCase):
    def test_grab_real_object(self):
        class Foo(object):
            pass
        foo = Foo()
        foo.bar = 'some string'
        assert h.grab(foo, 'bar') == 'some string'
        try:
            h.grab(foo, 'baz')
        except AttributeError:
            pass
        else:
            raise AssertionError

    def test_grab_any_fake(self):
        assert h.grab(None, 'nonexistent') == ''
        assert h.grab('', 'nonexistent') == ''

    def test_grab_date(self):
        assert h.grab(None, 'date') == date.today()

    def test_grab_user(self):
        class FakeRequest(object):
            pass
        class FakeUser(object):
            pass
        class SomeObject(object):
            pass
        req = FakeRequest()
        req.environ = {}
        req.environ['user'] = FakeUser()
        test_obj = SomeObject()
        req.environ['user'].id = test_obj
        request._push_object(req)
        assert h.grab(None, 'spender_id') == test_obj
        assert h.grab(None, 'creditor_id') == test_obj
        assert h.grab(None, 'debtor_id') == test_obj
        request._pop_object()

    def test_grab_amount(self):
        assert h.grab(None, 'amount') == 0



"""
Create subclasses that call out to their "superclass" for all methods
but return the "subclass's" type
"""

def wrapper(cls, func):
    return (lambda self, *args: cls(getattr(self.value, func)(*map(self.value.__class__, args))))

class SmartSubclass(object):
    def __init__(self, superclass, exclude=None):
        if exclude is None:
            exclude = []
        self.superclass = superclass
        self.exclude = exclude
    def __call__(self, name, bases, dict):
        c = type(name, bases, dict)
        for func in dir(self.superclass):
            if func not in dir(c) and \
                callable(getattr(self.superclass, func)) and \
                func not in self.exclude:
                setattr(c, func, wrapper(c, func))
        return c

__all__ = ['SmartSubclass']



from datetime import date
from formencode import Invalid

from webhelpers.pylonslib.secure_form import token_key

from bluechips.tests import *

from bluechips import model
from bluechips.model import meta
from bluechips.model.types import Currency

from bluechips.controllers.spend import ExpenditureSchema

class TestSpendController(TestController):

    def test_index(self):
        response = self.app.get(url_for(controller='spend'))
        # Test response...
        response.mustcontain('Add a New Expenditure')
        form = response.form

        user = meta.Session.query(model.User).\
                filter_by(name=u'Charlie Root').one()
        
        form['spender_id'] = user.id
        form['amount'] = '74.04'
        # Make sure date is today.
        today = date.today()
        assert form['date'].value == today.strftime('%m/%d/%Y')
        form['description'] = 'A test expenditure'
        form['shares-0.amount'] = '1'
        form['shares-1.amount'] = '2'
        form['shares-2.amount'] = '2'
        form['shares-3.amount'] = '1'
        for ii in range(4):
            assert int(form['shares-%d.user_id' % ii].value) == ii + 1

        response = form.submit()
        response = response.follow()
        response.mustcontain('Expenditure', 'created.')

        e = meta.Session.query(model.Expenditure).\
                order_by(model.Expenditure.id.desc()).first()
        assert e.spender.name == u'Charlie Root'
        assert e.amount == 7404
        assert e.date == today
        assert e.description == u'A test expenditure'

        # Test the split.
        shares = dict(((sp.user_id, sp.share)
                       for sp in e.splits))
        assert shares[1] == Currency('12.34')
        assert shares[2] == Currency('24.68')
        assert shares[3] == Currency('24.68')
        assert shares[4] == Currency('12.34')


    def test_edit_and_delete(self):
        user = meta.Session.query(model.User).\
                filter_by(name=u'Charlie Root').one()
        e = model.Expenditure(user, 53812, u'Lemon bundt cake', None)
        e.even_split()
        meta.Session.add(e)
        meta.Session.commit()

        response = self.app.get(url_for(controller='spend',
                                        action='edit',
                                        id=e.id))
        response.mustcontain('Edit an Expenditure')
        form = response.form

        assert int(form['spender_id'].value) == user.id
        assert form['amount'].value == '538.12'
        assert form['date'].value == date.today().strftime('%m/%d/%Y')
        assert form['description'].value == u'Lemon bundt cake'

        form['description'] = u'Updated bundt cake'

        # Update the split too.

        response = form.submit()
        response = response.follow()
        response.mustcontain('Expenditure', 'updated.')

        e = meta.Session.query(model.Expenditure).\
                order_by(model.Expenditure.id.desc()).first()
        assert e.description == u'Updated bundt cake'

        response = self.app.get(url_for(controller='spend',
                                        action='delete',
                                        id=e.id))
        response = response.form.submit('delete').follow()
        response.mustcontain('Expenditure', 'deleted')

    def test_delete_nonexistent(self):
        self.app.get(url_for(controller='spend',
                             action='delete',
                             id=124344),
                     status=404)

    def test_destroy_nonexistent(self):
        response = self.app.get(url_for(controller='spend',
                                        action='edit'))
        params = self.sample_params.copy()
        params[token_key] = response.form[token_key].value
        self.app.post(url_for(controller='spend',
                              action='destroy',
                              id=124344), 
                      params=params,
                      status=404)

    def test_delete_xsrf_protection(self):
        self.app.post(url_for(controller='spend',
                              action='destroy',
                              id=1),
                      params={'delete': 'Delete'},
                      status=403)

    def test_edit_zero_value(self):
        user = meta.Session.query(model.User).\
                filter_by(name=u'Charlie Root').one()
        e = model.Expenditure(user, 0, u'A zero value expenditure', None)
        e.even_split()
        meta.Session.add(e)
        meta.Session.commit()

        response = self.app.get(url_for(controller='spend',
                                        action='edit',
                                        id=e.id))
        response.mustcontain('Edit an Expenditure')
        form = response.form

        assert int(form['spender_id'].value) == user.id
        assert form['amount'].value == '0.00'
        assert form['date'].value == date.today().strftime('%m/%d/%Y')
        assert form['description'].value == u'A zero value expenditure'
        for ii in range(4):
            assert form['shares-%d.amount' % ii].value == '0'

    def test_edit_nonexistent(self):
        response = self.app.get(url_for(controller='spend',
                                        action='edit',
                                        id=124234), status=404)

    def test_update_nonexistent(self):
        response = self.app.get(url_for(controller='spend',
                                        action='edit'))
        params = self.sample_params.copy()
        params[token_key] = response.form[token_key].value
        self.app.post(url_for(controller='spend',
                              action='update',
                              id=14234), 
                      params=params,
                      status=404)

    def test_xsrf_protection(self):
        self.app.post(url_for(controller='spend',
                              action='update'),
                      params=self.sample_params,
                      status=403)

    def test_all_zero_shares_fails(self):
        params = self.sample_params.copy()
        for ii in range(4):
            params['shares-%d.amount' % ii] = '0'
        v = ExpenditureSchema()
        try:
            v.to_python(params)
        except Invalid:
            pass

    def setUp(self):
        self.sample_params = {
            'spender_id': '1',
            'amount': '44.12',
            'date': '10/5/2008',
            'description': 'Example expenditure post data.',
            'shares-0.user_id': '1',
            'shares-0.amount': '1',
            'shares-1.user_id': '2',
            'shares-1.amount': '1',
            'shares-2.user_id': '3',
            'shares-2.amount': '1',
            'shares-3.user_id': '4',
            'shares-3.amount': '1'}

    def tearDown(self):
        expenditures = meta.Session.query(model.Expenditure).all()
        for e in expenditures:
            meta.Session.delete(e)
        meta.Session.commit()



"""
Calculate the current state of the books
"""

import logging

from bluechips.lib.base import *
from bluechips.lib.totals import *

import sqlalchemy
from sqlalchemy import orm

from datetime import date, timedelta

from bluechips.model.types import Currency

from pylons import request

log = logging.getLogger(__name__)

class StatusController(BaseController):
    def index(self):
        c.debts = debts()
        c.settle = settle(c.debts)

        c.net = 0
        for from_user, to_user, amount in c.settle:
            if from_user == request.environ['user']:
                c.net -= amount
            elif to_user == request.environ['user']:
                c.net += amount
        
        periods = {}
        periods['Total'] = (None, None)
        periods['Past year'] = (date.today() - timedelta(days=365), None)
        periods['Year to date'] = (date.today().replace(month=1, day=1), None)
        periods['Month to date'] = (date.today().replace(day=1), None)
        periods['Last month'] = ((date.today() -
                                  timedelta(days=30)).replace(day=1),
                                 periods['Month to date'][0])
        
        c.totals = {}
        for period in periods.keys():
            c.totals[period] = {}
            start, end = periods[period]
            conds = []
            if start is not None:
                conds.append(model.Expenditure.date >= start)
            if end is not None:
                conds.append(model.Expenditure.date < end)
            if len(conds) > 1:
                conds = sqlalchemy.and_(*conds)
            elif len(conds) > 0:
                conds = conds[0]
            else:
                conds = None

            for scope in ('all', 'mine'):
                meth = getattr(self, '_total_%s' % scope)
                c.totals[period][scope] = meth(conds)

        c.expenditures = meta.Session.query(model.Expenditure).\
                filter(sqlalchemy.or_(
                    model.Expenditure.spender == request.environ['user'],
                    model.Expenditure.splits.any(
                        sqlalchemy.and_(
                            model.Split.user == request.environ['user'],
                            model.Split.share != 0)))).\
                options(orm.eagerload('splits')).\
                limit(10).all()
        c.transfers = meta.Session.query(model.Transfer).\
            filter(sqlalchemy.or_(
                model.Transfer.debtor==request.environ['user'],
                model.Transfer.creditor==request.environ['user'])).\
                limit(10).all()
        c.users = get_users()
        
        return render('/status/index.mako')
    
    def _total_all(self, conditions=None):
        q = meta.Session.query(sqlalchemy.func.SUM(
            model.Expenditure.amount))
        if conditions is not None:
            q = q.filter(conditions)
        return q.scalar()

    def _total_mine(self, conditions=None):
        q = meta.Session.query(sqlalchemy.func.SUM(
            model.Split.share)).join(model.Split.expenditure).\
                filter(model.Split.user == request.environ['user'])
        if conditions is not None:
            q = q.filter(conditions)
        return q.scalar()



"""
Calculate the current state of the books
"""

import logging

from bluechips.lib.base import *
from bluechips.lib.permissions import BlueChipResident

import sqlalchemy
from sqlalchemy import orm

from authkit.authorize.pylons_adaptors import authorize

from pylons import request
from pylons.decorators import validate
from pylons.decorators.secure import authenticate_form

from formencode import validators, Schema, FancyValidator, Invalid

log = logging.getLogger(__name__)


class EmailSchema(Schema):
    "Validate email updates."
    allow_extra_fields = False
    new_email = validators.Email()


class UniqueUsername(FancyValidator):
    def _to_python(self, value, state):
        u = meta.Session.query(model.User).\
            filter(model.User.username == value).\
            first()
        if u:
            raise Invalid(
                'That username already exists',
                value, state)
        return value


class NewUserSchema(Schema):
    "Validate new users."
    allow_extra_fields = False
    username = UniqueUsername(not_empty=True)
    password = validators.String(if_missing=None)
    confirm_password = validators.String(if_missing=None)
    name = validators.String(not_empty=False)
    resident = validators.StringBoolean(not_empty=True)
    chained_validators = [
        validators.FieldsMatch('password', 'confirm_password'),
        ]


class UserController(BaseController):
    def index(self):
        c.title = 'User Settings'
        return render('/user/index.mako')

    def email(self):
        c.title = 'User Settings'
        return render('/user/email.mako')

    @authenticate_form
    @validate(schema=EmailSchema(), form='index')
    def update(self):
        new_email = self.form_result['new_email']
        request.environ['user'].email = new_email
        meta.Session.commit()
        if new_email is None:
            h.flash("Removed email address.")
        else:
            h.flash("Updated email address to '%s'." % new_email)
        return h.redirect_to('/')

    @authorize(BlueChipResident())
    def new(self):
        c.title = 'Register a New User'
        return render('/user/new.mako')

    @authenticate_form
    @authorize(BlueChipResident())
    @validate(schema=NewUserSchema(), form='new')
    def create(self):
        u = model.User(username=self.form_result['username'],
                       resident=self.form_result['resident'])

        if self.form_result['name']:
            u.name = self.form_result['name']
        else:
            u.name = self.form_result['username']

        if self.form_result['password'] is not None:
            u.password = self.form_result['password']

        meta.Session.add(u)
        meta.Session.commit()

        h.flash('Successfully created new user %s' % u.username)
        return h.redirect_to('/')



"""Pylons environment configuration"""
import os

from mako.lookup import TemplateLookup
from pylons import config
from sqlalchemy import engine_from_config
from mailer import Mailer

import bluechips.lib.app_globals as app_globals
import bluechips.lib.helpers
from bluechips.config.routing import make_map
from bluechips.model import init_model

def load_environment(global_conf, app_conf):
    """Configure the Pylons environment via the ``pylons.config``
    object
    """
    # Pylons paths
    root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    paths = dict(root=root,
                 controllers=os.path.join(root, 'controllers'),
                 static_files=os.path.join(root, 'public'),
                 templates=[os.path.join(root, 'templates')])

    # Initialize config with the basic options
    config.init_app(global_conf, app_conf, package='bluechips', paths=paths)

    config['routes.map'] = make_map()
    config['pylons.app_globals'] = app_globals.Globals()
    config['pylons.h'] = bluechips.lib.helpers

    # Create the Mako TemplateLookup, with the default auto-escaping
    config['pylons.app_globals'].mako_lookup = TemplateLookup(
        directories=paths['templates'],
        module_directory=os.path.join(app_conf['cache_dir'], 'templates'),
        input_encoding='utf-8', output_encoding='utf-8',
        imports=['from webhelpers.html import escape'],
        default_filters=['escape'])
    
    # Setup SQLAlchemy database engine
    engine = engine_from_config(config, 'sqlalchemy.')
    init_model(engine)
    
    # CONFIGURATION OPTIONS HERE (note: all config options will override
    # any Pylons config options)
    config['pylons.app_globals'].mailer = Mailer(config.get('mailer.host',
                                                            '127.0.0.1'))



"""The application's Globals object"""

import logging

from pylons import config, request
from paste.deploy.converters import asbool
from mailer import Message

log = logging.getLogger(__name__)

class Globals(object):
    """Globals acts as a container for objects available throughout the
    life of the application
    """

    def __init__(self):
        """One instance of Globals is created during application
        initialization and is available during requests via the 'g'
        variable
        """
        pass

    def send_message(self, msg):
        """
        Wrap the call to mailer.send() so that we can do stuff like defer mail
        sending, wrap in a test fixture, selectively disable mailing in certain
        environments, etc.
        """
        if asbool(config.get('testing')) or asbool(config.get('network_free')):
            if 'mailer.messages' not in request.environ:
                request.environ['mailer.messages'] = []
            request.environ['mailer.messages'].append(msg)
            log.info("From: %s\nTo: %s\nSubject: %s\n\n%s",
                     msg.From, msg.To, msg.Subject, msg.Body)
        else:
            self.mailer.send(msg) # pragma: nocover

    def handle_notification(self, users, subject, body):
        "Send a notification email."
        recipients = [u.email for u in users if u.email is not None]
        if len(recipients) > 0:
            msg = Message(From=config.get('mailer.from',
                                          'root@localhost'),
                          To=recipients)
            msg.Subject = "BlueChips: %s" % subject
            msg.Body = body
            self.send_message(msg)



from pylons import config

from bluechips.tests import *
from bluechips import model
from bluechips.model import meta

class TestUserController(TestController):

    def test_email(self):
        response = self.app.get(url_for(controller='user', action='email'))
        # Test response...
        response.mustcontain('Email Notifications', 'User Settings')
        form = response.form
        form['new_email'] = 'test@example.com'
        response = form.submit().follow()
        response.mustcontain('Updated email address')

        user = meta.Session.query(model.User).\
                filter_by(username=unicode(config['fake_username'])).one()
        assert user.email == 'test@example.com'

    def test_clear_email(self):
        response = self.app.get(url_for(controller='user', action='email'))
        form = response.form
        form['new_email'] = ''
        response = form.submit().follow()
        response.mustcontain('Removed email address')

        user = meta.Session.query(model.User).\
                filter_by(username=unicode(config['fake_username'])).one()
        assert user.email == None

    def test_xsrf_protection(self):
        self.app.post(url_for(controller='user',
                              action='update'),
                      {'new_email': 'malicious@example.com'},
                      status=403)



from datetime import date
from decimal import Decimal

from webhelpers.pylonslib.secure_form import token_key

from bluechips.tests import *
from bluechips import model
from bluechips.model import meta

class TestTransferController(TestController):

    def test_index(self):
        response = self.app.get(url_for(controller='transfer'))
        # Test response...
        response.mustcontain('Add a New Transfer')
        form = response.form

        user_rich = meta.Session.query(model.User).\
                filter_by(name=u'Rich Scheme').one()
        user_ben = meta.Session.query(model.User).\
                filter_by(name=u'Ben Bitdiddle').one()

        form['debtor_id'] = user_rich.id
        form['creditor_id'] = user_ben.id
        form['amount'] = '123.45'
        # Make sure date is today.
        today = date.today()
        assert form['date'].value == today.strftime('%m/%d/%Y')
        form['description'] = 'A test transfer from Rich to Ben'

        response = form.submit()
        response = response.follow()
        response.mustcontain('Transfer', 'created.')

        t = meta.Session.query(model.Transfer).\
                order_by(model.Transfer.id.desc()).first()
        assert t.debtor.name == u'Rich Scheme'
        assert t.creditor.name == u'Ben Bitdiddle'
        assert t.amount == 12345
        assert t.date == today
        assert t.description == u'A test transfer from Rich to Ben'

    def test_edit_and_delete(self):
        user_rich = meta.Session.query(model.User).\
                filter_by(name=u'Rich Scheme').one()
        user_ben = meta.Session.query(model.User).\
                filter_by(name=u'Ben Bitdiddle').one()
        t = model.Transfer(user_rich, user_ben, 12345)
        t.description = u'Test transfer'
        meta.Session.add(t)
        meta.Session.commit()

        response = self.app.get(url_for(controller='transfer',
                                        action='edit',
                                        id=t.id))
        response.mustcontain('Edit a Transfer')
        form = response.form

        assert int(form['debtor_id'].value) == t.debtor_id
        assert int(form['creditor_id'].value) == t.creditor_id
        assert Decimal(form['amount'].value) * 100 == t.amount
        assert form['date'].value == t.date.strftime('%m/%d/%Y')
        assert form['description'].value == t.description

        form['description'] = u'A new description'

        response = form.submit()
        response = response.follow()
        response.mustcontain('Transfer', 'updated.')

        t = meta.Session.query(model.Transfer).\
                order_by(model.Transfer.id.desc()).first()
        assert t.description == u'A new description'

        response = self.app.get(url_for(controller='transfer',
                                        action='delete',
                                        id=t.id))
        response = response.form.submit('delete').follow()
        response.mustcontain('Transfer', 'deleted')

    def test_edit_nonexistent(self):
        response = self.app.get(url_for(controller='transfer',
                                        action='edit',
                                        id=21424), status=404)

    def test_update_nonexistent(self):
        response = self.app.get(url_for(controller='transfer',
                                        action='edit'))
        params = self.sample_params.copy()
        params[token_key] = response.form[token_key].value
        self.app.post(url_for(controller='transfer',
                              action='update',
                              id=21424),
                      params=params,
                      status=404)

    def test_xsrf_protection(self):
        self.app.post(url_for(controller='transfer',
                              action='update'),
                      params=self.sample_params,
                      status=403)


    def test_update_get_redirects(self):
        response = self.app.get(url_for(controller='transfer',
                                        action='update'),
                                status=302)
        assert (dict(response.headers)['location'] ==
                url_for(controller='transfer', action='edit', qualified=True))

    def test_delete_nonexistent(self):
        self.app.get(url_for(controller='transfer',
                             action='delete',
                             id=124244),
                     status=404)

    def test_destroy_nonexistent(self):
        response = self.app.get(url_for(controller='transfer',
                                        action='edit'))
        params = self.sample_params.copy()
        params[token_key] = response.form[token_key].value
        self.app.post(url_for(controller='transfer',
                              action='destroy',
                              id=124344), 
                      params=params,
                      status=404)

    def test_delete_xsrf_protection(self):
        self.app.post(url_for(controller='transfer',
                              action='destroy',
                              id=1),
                      params={'delete': 'Delete'},
                      status=403)

    def setUp(self):
        self.sample_params = {
            'debtor_id': '1',
            'creditor_id': '2',
            'amount': '33.98',
            'date': '4/1/2007',
            'description': 'Example transfer params.'}

    def tearDown(self):
        transfers = meta.Session.query(model.Transfer).all()
        for t in transfers:
            meta.Session.delete(t)
        meta.Session.commit()



from bluechips.model.user import User
from bluechips.model.split import Split
from bluechips.model import meta
from bluechips.model.types import Currency
from decimal import Decimal
from datetime import datetime
import random

class Expenditure(object):
    def __init__(self, spender=None, amount=Currency(0), description=u"",
                 date=None):
        self.spender = spender
        self.amount = amount
        self.description = description
        if self.date == None:
            self.date = datetime.now()
    
    def __repr__(self):
        return '<Expenditure: spender: %s spent: %s>' % (self.spender,
                                                         self.amount)

    def even_split(self):
        """
        Split up an expenditure evenly among the resident users
        """
        
        residents = meta.Session.query(User).filter(User.resident==True)
        split_percentage = Decimal(100) / Decimal(residents.count())
        self.split(dict((resident, split_percentage) for resident in residents))
    
    def split(self, split_dict):
        """
        Split up an expenditure.
        
        split_dict should be a dict mapping from bluechips.model:User
        objects to a decimal:Decimal object representing the percentage
        that user is responsible for.
        
        Percentages will be normalized to sum to 100%.
        
        If the split leaks or gains money due to rounding errors, the
        pennies will be randomly distributed to one of the users.
        
        I mean, come on. You're already living together. Are you really
        going to squabble over a few pennies?
        """
        
        map(meta.Session.delete, meta.Session.query(Split).\
                filter_by(expenditure_id=self.id))
        
        total = sum(split_dict.itervalues())
        
        for user, share in split_dict.items():
            if share == 0:
                del split_dict[user]
            
        amounts_dict = dict()
        
        for user, share in split_dict.iteritems():
            amounts_dict[user] = Currency((share * self.amount) / total)
        
        difference = self.amount - sum(amounts_dict.itervalues())
        
        if difference > 0:
            for i in xrange(difference):
                winner = random.choice(amounts_dict.keys())
                amounts_dict[winner] += Currency(1)
        elif difference < 0:
            for i in xrange(-difference):
                winner = random.choice(amounts_dict.keys())
                amounts_dict[winner] -= Currency(1)
        
        for user, share in amounts_dict.iteritems():
            s = Split(self, user, share)
            meta.Session.add(s)

    def involves(self, user):
        "Returns True if ``user`` is involved in this expenditure."
        return (any((split.user == user) and (split.share != 0)
                    for split in self.splits) or
                (self.spender == user))

    def share(self, user):
        "Return the share corresponding to ``user``."
        shares = dict((split.user, split.share)
                      for split in self.splits)
        return shares.get(user, Currency(0))

__all__ = ['Expenditure']



#!/usr/bin/python

try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

setup(
    name='BlueChips',
    version='1.0.1',
    description='BlueChips - finances for people with shared expenses',
    long_description=open('README.rst').read(),
    author='Residents of Blue Sun Corporate Headquarters',
    author_email='chips@blue-sun-corp.com',
    url='http://github.com/ebroder/bluechips',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Framework :: Pylons',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Natural Language :: English',
        'Topic :: Home Automation',
        'Topic :: Internet :: WWW/HTTP :: WSGI :: Application',
        'Topic :: Office/Business :: Financial :: Accounting',
        ],
    install_requires=["Pylons>=0.9.6",
                      "WebHelpers>=1.0",
                      "SQLAlchemy>=0.5",
                      "AuthKit>=0.4.0",
                      "FormEncode>=1.2.1",
                      "mailer>=0.5"],
    setup_requires=["PasteScript==dev,>=1.6.3dev-r7326"],
    packages=find_packages(exclude=['ez_setup']),
    include_package_data=True,
    test_suite='nose.collector',
    zip_safe=False,
    paster_plugins=['PasteScript', 'Pylons'],
    entry_points="""
    [paste.app_factory]
    main = bluechips.config.middleware:make_app

    [paste.app_install]
    main = pylons.util:PylonsInstaller
    """,
)



"""
Define special types used in BlueChips
"""

import locale
from decimal import Decimal, InvalidOperation

import sqlalchemy as sa
from formencode import validators, Invalid
from bluechips.lib.subclass import SmartSubclass

from weakref import WeakValueDictionary

def localeconv():
    "Manually install en_US for systems that don't have it."
    d = {'currency_symbol': '$',
     'decimal_point': '.',
     'frac_digits': 2,
     'grouping': [3, 3, 0],
     'int_curr_symbol': 'USD ',
     'int_frac_digits': 2,
     'mon_decimal_point': '.',
     'mon_grouping': [3, 3, 0],
     'mon_thousands_sep': ',',
     'n_cs_precedes': 1,
     'n_sep_by_space': 0,
     'n_sign_posn': 1,
     'negative_sign': '-',
     'p_cs_precedes': 1,
     'p_sep_by_space': 0,
     'p_sign_posn': 1,
     'positive_sign': '',
     'thousands_sep': ','}
    return d
locale.localeconv = localeconv


class CurrencyValidator(validators.FancyValidator):
    "A validator to convert to Currency objects."
    messages = {'amount': "Please enter a valid currency amount",
                'precision': "Only two digits after the decimal, please",
                'nonzero': "Please enter a non-zero amount"}

    def _to_python(self, value, state):
        try:
            dec = Decimal(value)
        except InvalidOperation:
            raise Invalid(self.message('amount', state),
                          value, state)
        else:
            ret = dec.quantize(Decimal('1.00'))
            if ret == 0:
                raise Invalid(self.message('nonzero', state),
                              value, state)
            elif ret != dec:
                raise Invalid(self.message('precision', state),
                              value, state)
            else:
                return Currency(int(ret * 100))


class Currency(object):
    """
    Store currency values as an integral number of cents
    """
    __metaclass__ = SmartSubclass(int)
    __old_values__ = WeakValueDictionary()
    def __new__(cls, value):
        if value is None:
            value = 0
        elif isinstance(value, str):
            value = int(float(value) * 100)
        else:
            value = int(value)
        
        if value not in cls.__old_values__:
            new_object = super(cls, cls).__new__(cls)
            new_object.value = value
            cls.__old_values__[value] = new_object
            return new_object
        else:
            return cls.__old_values__[value]
    
    def __int__(self):
        """
        If I don't define this, SmartSubclass will return
        Currency(int(self.value))
        """
        return self.value
    def __float__(self):
        """
        If I don't define this, SmartSubclass will return
        Currency(float(self.value))
        """
        return float(self.value)
    def __long__(self):
        """
        If I don't define this, SmartSubclass will return
        Currency(long(self.value))
        """
        return long(self.value)
    
    def __cmp__(self, other):
        """
        This is overridden for when validators compare a Currency to
        ''
        """
        if other == '':
            return 1
        else:
            return self.value.__cmp__(int(other))
    
    def __mul__(self, other):
        """
        If I don't define this, SmartSubclass will convert the other
        argument to an int
        """
        return Currency(self.value * other)
    def __rmul__(self, other):
        """
        If I don't define this, SmartSubclass will convert the other
        argument to an int
        """
        return self.__mul__(other)
    def __div__(self, other):
        """
        If I don't define this, SmartSubclass will convert the other
        argument to an int
        """
        return Currency(self.value / other)
    def __truediv__(self, other):
        """
        If I don't define this, SmartSubclass will convert the other
        argument to an int
        """
        return Currency(self.value / other)
    
    def __repr__(self):
        return '%s("%s")' % (self.__class__.__name__, str(self))
    def __str__(self):
        return locale.currency(self.value / 100., grouping=True)


class DBCurrency(sa.types.TypeDecorator):
    """
    A type which represents monetary amounts internally as integers.
    
    This avoids binary/decimal float conversion issues
    """
    
    impl = sa.types.Integer
    
    def process_bind_param(self, value, engine):
        return int(value)
    
    def convert_result_value(self, value, engine):
        return Currency(value)
    process_result_value = convert_result_value



from unittest import TestCase
from bluechips.tests import *
from bluechips import model
from bluechips.model import meta
from webhelpers.number import standard_deviation as std_dev

class TestSplitRandom(TestCase):
    @classmethod
    def setUpClass(cls):
        createUsers()
    
    @classmethod
    def tearDownClass(cls):
        deleteUsers()
    
    def setUp(self):
        createExpenditures()
    
    def tearDown(self):
        deleteExpenditures()
    
    def test_splitTotal(self):
        for e in meta.Session.query(model.Expenditure):
            self.assertEqual(sum(s.share for s in e.splits), e.amount)
    
    def test_splitDistribution(self):
        user_count = meta.Session.query(model.User).count()
        for e in meta.Session.query(model.Expenditure):
            even_total = (e.amount / user_count) * user_count
            difference = abs(even_total - e.amount)
            self.assert_(std_dev(list(int(s.share) for s in e.splits)) <= difference, \
                "Expenditure doesn't appear to be evenly distributed")



"""SQLAlchemy Metadata and Session object"""
from sqlalchemy import MetaData
from sqlalchemy.orm import scoped_session, sessionmaker

# SQLAlchemy database engine.  Updated by model.init_model()
engine = None

# SQLAlchemy session manager.  Updated by model.init_model()
Session = None

# Global metadata. If you have multiple databases with overlapping table
# names, you'll need a metadata for each database
metadata = MetaData()

__all__ = ['engine', 'Session', 'metadata']



from types import Currency

class Subitem(object):
    def __init__(self, expenditure=None, user=None, amount=Currency(0)):
        self.expenditure = expenditure # pragma: nocover
        self.user = user # pragma: nocover
        self.share = share # pragma: nocover
        
    def __repr__(self):
        return ('<Subitem: expense: %s user: %s cost: %s>' %
                (self.expense, self.user, self.amount)) # pragma: nocover

__all__ = ['Subitem']



"""Setup the BlueChips application"""
import logging

from bluechips.config.environment import load_environment

log = logging.getLogger(__name__)

def setup_app(command, conf, vars):
    """Place any commands to setup bluechips here"""
    load_environment(conf.global_conf, conf.local_conf)

    # Load the models
    from bluechips.model import meta
    meta.metadata.bind = meta.engine

    # Create the tables if they aren't there already
    meta.metadata.create_all(checkfirst=True)



"""The application's model objects"""
import sqlalchemy as sa
from sqlalchemy import orm

from bluechips.model.user import User
from bluechips.model.expenditure import Expenditure
from bluechips.model.split import Split
from bluechips.model.subitem import Subitem
from bluechips.model.transfer import Transfer

from bluechips.model import meta
from bluechips.model import types

from datetime import datetime

def init_model(engine):
    """Call me before using any of the tables or classes in the model"""

    sm = orm.sessionmaker(autoflush=True, bind=engine)

    meta.engine = engine
    meta.Session = orm.scoped_session(sm)

### Database Schemas ###

users = sa.Table('users', meta.metadata,
                 sa.Column('id', sa.types.Integer, primary_key=True),
                 sa.Column('username', sa.types.Unicode(32), nullable=False),
                 sa.Column('name', sa.types.Unicode(64)),
                 sa.Column('resident', sa.types.Boolean, default=True),
                 sa.Column('email', sa.types.Unicode(64)),
                 sa.Column('password', sa.types.Unicode(64)),
                 )

expenditures = sa.Table('expenditures', meta.metadata,
                        sa.Column('id', sa.types.Integer, primary_key=True),
                        sa.Column('spender_id', sa.types.Integer,
                                  sa.ForeignKey('users.id'), nullable=False),
                        sa.Column('amount', types.DBCurrency, nullable=False),
                        sa.Column('description', sa.types.Text),
                        sa.Column('date', sa.types.Date, default=datetime.now),
                        sa.Column('entered_time', sa.types.DateTime, 
                                  default=datetime.utcnow)
                        )

splits = sa.Table('splits', meta.metadata,
                  sa.Column('id', sa.types.Integer, primary_key=True),
                  sa.Column('expenditure_id', sa.types.Integer,
                            sa.ForeignKey('expenditures.id'), nullable=False),
                  sa.Column('user_id', sa.types.Integer,
                            sa.ForeignKey('users.id'), nullable=False),
                  sa.Column('share', types.DBCurrency, nullable=False)
                  )

subitems = sa.Table('subitems', meta.metadata,
                    sa.Column('id', sa.types.Integer, primary_key=True),
                    sa.Column('expenditure_id', sa.types.Integer,
                              sa.ForeignKey('expenditures.id'), nullable=False),
                    sa.Column('user_id', sa.types.Integer,
                              sa.ForeignKey('users.id'), nullable=False),
                    sa.Column('amount', types.DBCurrency, nullable=False)
                    )

transfers = sa.Table('transfers', meta.metadata,
                     sa.Column('id', sa.types.Integer, primary_key=True),
                     sa.Column('debtor_id', sa.types.Integer,
                               sa.ForeignKey('users.id'), nullable=False),
                     sa.Column('creditor_id', sa.types.Integer,
                               sa.ForeignKey('users.id'), nullable=False),
                     sa.Column('amount', types.DBCurrency, nullable=False),
                     sa.Column('description', sa.Text, default=None),
                     sa.Column('date', sa.types.Date, default=datetime.now),
                     sa.Column('entered_time', sa.types.DateTime,
                               default=datetime.utcnow)
                     )

### DB/Class Mapping ###

orm.mapper(User, users,
           properties={
        'expenditures': orm.relation(Expenditure,
                                     backref='spender')
})

orm.mapper(Expenditure, expenditures,
           order_by=[expenditures.c.date.desc(), expenditures.c.entered_time.desc()],
           properties={
        'splits': orm.relation(Split, backref='expenditure',
                               cascade='all, delete'),
        'subitems': orm.relation(Subitem, backref='expenditure',
                                 cascade='all, delete')
})

orm.mapper(Split, splits, properties={
        'user': orm.relation(User)
})

orm.mapper(Subitem, subitems, properties={
        'user': orm.relation(User)
})

orm.mapper(Transfer, transfers,
           order_by=[transfers.c.date.desc(), transfers.c.entered_time.desc()],
           properties={
        'debtor': orm.relation(User,
                               primaryjoin=(transfers.c.debtor_id==\
                                                users.c.id)),
        'creditor': orm.relation(User,
                                 primaryjoin=(transfers.c.creditor_id==\
                                                  users.c.id))
})

__all__ = ['users', 'expenditures', 'splits', 'subitems', 'transfers',
           'User', 'Expenditure', 'Split', 'Subitem', 'Transfer',
           'meta']



from unittest import TestCase
from bluechips.lib import totals

class TestReorderingSettle(TestCase):
    def test_transfer_minimized(self):
        """
        Test that the number of transfers is actually minimized.

        Taken from a real-world situation, we discovered that failing
        to re-order the debt lists after every transfer could lead to
        extra, unnecessary transfers.
        """
        self.assertEqual(len(totals.settle({'Alice': 100,
                                            'Bob': -85,
                                            'Charlie': 35,
                                            'Dave': -35,
                                            'Eve': -15})),
                         3)

    def test_settle_even(self):
        transfers = totals.settle({'Alice': 0,
                                   'Bob': 0,
                                   'Charlie': 0})
        assert transfers == []
    
    def test_settle_positive(self):
        transfers = totals.settle({'Alice': -50,
                                   'Bob': 100,
                                   'Charlie': -50})
        assert transfers == [('Bob', 'Charlie', 50),
                             ('Bob', 'Alice', 50)]

    def test_settle_uneven_positive(self):
        try:
            transfers = totals.settle({'Alice': -50,
                                       'Bob': -50,
                                       'Charlie': -50})
        except totals.DirtyBooks:
            pass

    def test_settle_uneven_negative(self):
        try:
            transfers = totals.settle({'Alice': 50,
                                       'Bob': 50,
                                       'Charlie': 50})
        except totals.DirtyBooks:
            pass
